package sorting;

import java.util.ArrayList;
import java.util.Stack;

/**
 * A class that contains several sorting routines, implemented as static
 * methods. Arrays are rearranged with smallest item first, using compareTo.
 * Show intermediate results with annotation.
 * 
 * @author Mark Allen Weiss (modified by Ron Mak)
 */
public final class ShowSort 
{
    /**
     * Simple insertion sort.
     * @param a an array of Comparable items.
     */
    public static <AnyType extends Comparable<? super AnyType>> 
        void insertionSort(AnyType[] a) 
    {
        int j;

        for (int pass = 1; pass < a.length; pass++) {
            int moves = 0;
            AnyType tmp = a[pass];
            ArrayList<Integer> p = new ArrayList<>();
            
            for (j = pass; j >= 0; j--) p.add(j);
            print(a, p);
            
            for (j = pass; j > 0 && tmp.compareTo(a[j-1]) < 0; j--) {
                a[j] = a[j-1];
                moves++;
            }
            
            if (pass != j) {
                a[j] = tmp;
                moves++;
            }
            
            System.out.printf(" (%d moves)\n", moves);
        }
        
        print(a, null);
        System.out.println(" (sorted)");
    }

    /**
     * Shellsort using Knuth's increment sequence for h.
     * @param a an array of Comparable items to sort.
     */
    public static <AnyType extends Comparable<? super AnyType>> 
        void shellsort(AnyType[] a) 
    {
        int j;

        // Compute highest h value.
        int h = 1;
        while (h < a.length) h = 3*h + 1; 
        
        for (h = (h-1)/3; h >= 1; h = (h-1)/3) {
            System.out.printf("\nh = %d\n\n", h);
            
            for (int pass = h; pass < a.length; pass++) {
                int moves = 0;
                AnyType tmp = a[pass];
                ArrayList<Integer> p = new ArrayList<>();
                
                for (j = pass; j >= 0; j -= h) p.add(j);
                print(a, p);
                
                for (j = pass; j >= h && tmp.compareTo(a[j-h]) < 0; j -= h) {
                    a[j] = a[j-h];
                    moves++;
                }
                
                if (pass != j) {
                    a[j] = tmp;
                    moves++;
                }
                
                System.out.printf(" (%d moves)\n", moves);
            }
        }
        
        print(a, null);
        System.out.println(" (sorted)");
    }

    /**
     * Standard heapsort.
     * @param a an array of Comparable items.
     */
    public static <AnyType extends Comparable<? super AnyType>> 
        void heapsort(AnyType[] a) 
    {
        // Build heap.
        for (int i = a.length/2 - 1; i >= 0; i--) {
            percDown(a, i, a.length);
        }
        
        // Delete max.
        for (int i = a.length - 1; i > 0; i--) {
            ArrayList<Integer> p = new ArrayList<>();
            p.add(0);
            p.add(i);
            print(a, p);
            System.out.println(" (swap, 2 moves)");
            
            swapReferences(a, 0, i);
            percDown(a, 0, i);
        }
        
        print(a, null);
        System.out.println(" (sorted)");
    }

    /**
     * Internal method for heapsort that is used in deleteMax and buildHeap.
     * @param a an array of Comparable items.
     * @index i the position from which to percolate down.
     * @int n the logical size of the binary heap.
     */
    private static <AnyType extends Comparable<? super AnyType>> 
        void percDown(AnyType[] a, int i, int n) 
    {
        int child;
        int iSave = i;
        int moves = 0;
        AnyType tmp;
        
        ArrayList<Integer> p = new ArrayList<>();
        p.add(i);
        print(a, null);
        System.out.println();
        
        for (tmp = a[i]; leftChild(i) < n; i = child) {
            child = leftChild(i);
            
            if ((child != n-1) && (a[child].compareTo(a[child+1]) < 0)) {
                child++;
            }
            
            if (tmp.compareTo(a[child]) < 0) {
                a[i] = a[child];
                p.add(child);
                moves++;
            }
            else break;
        }
        
        if (i != iSave) {
            a[i] = tmp;
            moves++;
        }
        
        print(p, a.length);
        System.out.printf(" (percolate down, %d moves)\n", moves);
    }

    /**
     * Internal method for heapsort.
     * @param i the index of an item in the heap.
     * @return the index of the left child.
     */
    private static int leftChild(int i) 
    {
        return 2*i + 1;
    }

    /**
     * Mergesort algorithm.
     * @param a an array of Comparable items to sort.
     */
    public static <AnyType extends Comparable<? super AnyType>> 
        void mergeSort(AnyType[] a) 
    {
        AnyType[] tmpArray = (AnyType[]) new Comparable[a.length];
        mergeSort(a, tmpArray, 0, a.length - 1);
        
        print(a, null);
        System.out.println(" (sorted)");
    }

    /**
     * Internal method for mergesort that makes recursive calls.
     * @param a an array of Comparable items.
     * @param tmpArray an array to place the merged result.
     * @param left the left-most index of the subarray.
     * @param right the right-most index of the subarray.
     */
    private static <AnyType extends Comparable<? super AnyType>> 
        void mergeSort(AnyType[] a, AnyType[] tmpArray, int left, int right) 
    {
        if (left < right) {
            ArrayList<Integer> p = new ArrayList<>();
            
            for (int k = left; k <= right; k++) p.add(k);
            print(a, p);
            
            int center = (left + right)/2;
            System.out.printf(" (center = %d)\n", a[center]);
            
            mergeSort(a, tmpArray, left, center);
            mergeSort(a, tmpArray, center+1, right);
            merge(a, tmpArray, left, center+1, right);
            
            for (int k = left; k <= right; k++) p.add(k);
            print(a, p);
            System.out.println(" (merged)");
        }
    }

    /**
     * Internal method for mergesort that merges two sorted subarray halves.
     * @param a an array of Comparable items.
     * @param tmpArray an array to place the merged result.
     * @param leftPos the left-most index of the subarray.
     * @param rightPos the index of the start of the second half.
     * @param rightEnd the right-most index of the subarray.
     */
    private static <AnyType extends Comparable<? super AnyType>> 
        void merge(AnyType[] a, AnyType[] tmpArray, 
                   int leftPos, int rightPos, int rightEnd) 
    {
        int leftEnd = rightPos - 1;
        int tmpPos = leftPos;
        int numElements = rightEnd - leftPos + 1;

        // Main loop
        while (leftPos <= leftEnd && rightPos <= rightEnd) {
            if (a[leftPos].compareTo(a[rightPos]) <= 0) {
                tmpArray[tmpPos++] = a[leftPos++];
            }
            else {
                tmpArray[tmpPos++] = a[rightPos++];
            }
        }

        // Copy rest of first half.
        while (leftPos <= leftEnd) {
            tmpArray[tmpPos++] = a[leftPos++];
        }

        // Copy rest of right half.
        while (rightPos <= rightEnd) {
            tmpArray[tmpPos++] = a[rightPos++];
        }

        // Copy tmpArray back.
        for (int i = 0; i < numElements; i++, rightEnd--) {
            a[rightEnd] = tmpArray[rightEnd];
        }
    }

    /**
     * Quicksort algorithm.
     * @param a an array of Comparable items to sort.
     */
    public static <AnyType extends Comparable<? super AnyType>> 
        void quicksort(AnyType[] a) 
    {
        quicksort(a, 0, a.length - 1);
        
        print(a, null);
        System.out.println(" (sorted)");
    }

    private static final int CUTOFF = 3;

    /**
     * Method to swap to elements in an array.
     * @param a an array of objects.
     * @param index1 the index of the first object.
     * @param index2 the index of the second object.
     */
    private static <AnyType> void swapReferences(AnyType[] a, int index1,
                                                              int index2) 
    {
        AnyType tmp = a[index1];
        a[index1] = a[index2];
        a[index2] = tmp;
    }

    /**
     * Return median of left, center, and right. Order these and hide the pivot.
     */
    private static <AnyType extends Comparable<? super AnyType>> 
        AnyType median3(AnyType[] a, int left, int right) 
    {
        int center = (left + right)/2;
        System.out.printf(" (%d:%d:%d", a[left], a[center], a[right]);
        
        if (a[center].compareTo(a[left]) < 0) {
            swapReferences(a, left, center);
        }
        if (a[right].compareTo(a[left]) < 0) {
            swapReferences(a, left, right);
        }
        if (a[right].compareTo(a[center]) < 0) {
            swapReferences(a, center, right);
        }

        // Place the pivot at position right-1.
        swapReferences(a, center, right-1);
        
        System.out.printf(" pivot = %d)\n", a[right-1]);
        return a[right-1];
    }
    
    /**
     * Internal quicksort method that makes recursive calls. Uses
     * median-of-three partitioning and a cutoff of 10.
     * @param a an array of Comparable items.
     * @param left the left-most index of the subarray.
     * @param right the right-most index of the subarray.
     */
    private static <AnyType extends Comparable<? super AnyType>> 
        void quicksort(AnyType[] a, int left, int right) 
    {
        ArrayList<Integer> p = new ArrayList<>();
        for (int k = left; k <= right; k++) p.add(k);
        
        if (left + CUTOFF <= right) {
            print(a, p);
            AnyType pivot = median3(a, left, right);

            // Begin partitioning
            int i = left, j = right-1;
            for (;;) {
                while (a[++i].compareTo(pivot) < 0);
                while (a[--j].compareTo(pivot) > 0);
                
                if (i < j) {
                    swapReferences(a, i, j);
                }
                else break;
            }

            swapReferences(a, i, right-1); // Restore pivot

            quicksort(a, left, i-1);       // Sort small elements
            quicksort(a, i+1, right);      // Sort large elements
        } 
        else {
            // Do an insertion sort on the subarray
            print(a, p);
            insertionSort(a, left, right);
            System.out.println(" (insertion sort)");
        }
    }

    /**
     * Internal insertion sort routine for subarrays that is used by quicksort.
     * 
     * @param a an array of Comparable items.
     * @param left the left-most index of the subarray.
     * @param right the right-most index of the subarray.
     */
    private static <AnyType extends Comparable<? super AnyType>> 
        void insertionSort( AnyType[] a, int left, int right) 
    {
        for (int p = left + 1; p <= right; p++) {
            AnyType tmp = a[p];
            int j;

            for (j = p; j > left && tmp.compareTo(a[j - 1]) < 0; j--) {
                a[j] = a[j - 1];
            }
            
            a[j] = tmp;
        }
    }
    
    /**
     * Print an array with position markers.
     * @param a the array to print.
     * @param p the array of positions to mark.
     */
    private static <AnyType extends Comparable<? super AnyType>> 
        void print(AnyType[] a, ArrayList<Integer> p)
    {
        // Print the array.
        for (AnyType value : a) {
            System.out.printf("%3d", value);
        }

        // Print the markers.
        if (p != null) {
            System.out.println();
            print(p, a.length);
        }
    }
    
    /**
     * Print the position markers.
     * @param n the length of the array to mark.
     */
    private static void print(ArrayList<Integer> p, int length)
    {
        for (int i = 0; i < length; i++) {
            System.out.print(p.contains(i) ? " ^^" : "   ");
        }
    }
    
    /**
     * Generate an array of n unique random integers.
     * @param n the value of n.
     * @return the array of unique random integers.
     */
    private static Integer[] uniqueRandom(int n)
    {
        int i = 0;
        java.util.Random r = new java.util.Random();
        ArrayList<Integer> list = new ArrayList<>();
        Integer a[] = new Integer[n];
        
        while (i < n) {
            int k = r.nextInt(90) + 10;  // random value 10..99;
            
            // Append the value to the array 
            // if it isn't already in there.
            if (!list.contains(k)) {
                list.add(k);
                a[i++] = k;
            }
        }
        
        return a;
    }
    
    /**
     * Make a copy of an array of Integer.
     * @param a the array to copy.
     * @return the copy.
     */
    private static Integer[] copyOf(Integer a[])
    {
        Integer copy[] = new Integer[a.length];
        for (int i = 0; i < a.length; i++) copy[i] = a[i];
        return copy;
    }

    /**
     * Main.
     * @param args the command-line arguments (ignored).
     */
    public static void main(String[] args) 
    {     
        final int N = 25;
        Integer unsorted[] = uniqueRandom(N);
        
        System.out.println("Insertion sort\n");
        ShowSort.insertionSort(copyOf(unsorted));
        
        System.out.println("\nShellsort");
        ShowSort.shellsort(copyOf(unsorted));
        
        System.out.println("\nHeap sort\n");
        ShowSort.heapsort(copyOf(unsorted));
        
        System.out.println("\nMerge sort\n");
        ShowSort.mergeSort(copyOf(unsorted));
        
        System.out.println("\nQuicksort\n");
        ShowSort.quicksort(copyOf(unsorted));
    }
}
